#if defined(_WIN32)

/* Windows */
# include <windows.h>
# include <tchar.h>

#else

/* Linux or VxWorks */
# define _T(x) x
# define _tprintf printf
# define _tcsicmp(pStr1, pStr2) strcasecmp(pStr1, pStr2)

#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#if !defined(_MSC_VER)
/* For strcasecmp in POSIX-compliant environment */
# include <strings.h>
#endif

#include <sdk_common.h>
#include <platform_time.h>

#include "exit_code.h"
#include "config_si5338.h"
#include "si5338_regmap.h"

/* Register addresses in SI5338 */
#define SI5338_REGIDX_REV_ID           (0U)
#define SI5338_REGIDX_DEV_CONFIG2      (2U)
#define SI5338_REGIDX_DEV_CONFIG3      (3U)
#define SI5338_REGIDX_DEV_CONFIG4      (4U)
#define SI5338_REGIDX_DEV_CONFIG5      (5U)
#define SI5338_REGIDX_FCAL_OVRD0      (45U)
#define SI5338_REGIDX_FCAL_OVRD1      (46U)
#define SI5338_REGIDX_FCAL_OVRD2      (47U)
#define SI5338_REGIDX_FCAL_OVRD_EN    (49U)
#define SI5338_REGIDX_LOL_STATUS     (218U)
#define SI5338_REGIDX_DRIVER_DISABLE (230U)
#define SI5338_REGIDX_FCAL0          (235U)
#define SI5338_REGIDX_FCAL1          (236U)
#define SI5338_REGIDX_FCAL2          (237U)
#define SI5338_REGIDX_LOL_PAUSE      (241U)
#define SI5338_REGIDX_PLL_SOFT_RESET (246U)
#define SI5338_REGIDX_PAGE           (255U)

static void messageCallbackRoutine(
  const TCHAR* pszMessage,
  SI5338RegMapMessageType nMessageType)
{
  const TCHAR* pszMessageType = _T("");

  switch (nMessageType) {
  case SI5338RegMapMessageInfo:
    pszMessageType = _T("Info");
    break;

  case SI5338RegMapMessageWarning:
    pszMessageType = _T("Warning");
    break;

  case SI5338RegMapMessageError:
    pszMessageType = _T("Error");
    break;
  }

  _tprintf(_T("%s: %s\n"), pszMessageType, pszMessage);
}

static void dumpRegMap(
  const SI5338RegMap* pRegMap)
{
  unsigned int i, n = pRegMap->nEntryCount;

  _tprintf(_T("%u register map entries(s):\n"), n);

  for (i = 0; i < n; i++) {
    _tprintf(
      _T("#%03u index=0x%02X%02X mask=0x%02X value=0x%02X\n"),
      i, pRegMap->entries[i].nRegIndexHi, pRegMap->entries[i].nRegIndexLo, pRegMap->entries[i].nMaskValue, pRegMap->entries[i].nRegValue);
  }
}

static int
readRegInner(
  const DeviceHandle* pDevice,
  bool_t bVerbose,
  uint8_t nBusIndex,
  uint8_t nDevIndex,
  uint8_t nRegIndex,
  uint8_t* pnData)
{
  uint8_t command[6];
  uint8_t response[4];
  uint8_t expResponse[2];
  uint32_t responseLength;
  uint8_t avr2Status;
  int ret = EXIT_OK;

  command[0] = 0x05; /* User extended command group */
  command[1] = 0x80; /* I2CRead command */
  command[2] = nBusIndex; /* I2C bus number */
  command[3] = nDevIndex; /* I2C device number */
  command[4] = nRegIndex; /* Address within device */
  command[5] = 1; /* Number of bytes to read, returned in response */
  expResponse[0] = 0x05; /* User extended command group */
  expResponse[1] = 0x00; /* I2CRead response */

  memset(response, 0, sizeof(response));
  ret = avr2utilSendAvr2Command(
    pDevice,
    bVerbose,
    0,
    command,
    sizeof(command),
    expResponse,
    response,
    sizeof(response),
    &responseLength);
  if (EXIT_OK != ret) {
    goto done;
  }

  avr2Status = response[2];
  if (0 != avr2Status) {
    const TCHAR* pReason;

    switch (avr2Status) {
    case 1:
      pReason = _T("General failure");
      break;

    default:
      pReason = _T("** UNEXPECTED ERROR CODE **");
      break;
    }
    _tprintf(_T("*** AVR2 returned status 0x%02X => %s\n"),
      (unsigned int)response[2], pReason);
    ret = EXIT_AVR2_BAD_STATUS;
    goto done;
  }

  *pnData = response[3];

done:
  return ret;
}

static int
writeRegInner(
  const DeviceHandle* pDevice,
  bool_t bVerbose,
  uint8_t nBusIndex,
  uint8_t nDevIndex,
  uint8_t nRegIndex,
  uint8_t nData)
{
  uint8_t command[7];
  uint8_t response[3];
  uint8_t expResponse[2];
  uint32_t responseLength;
  uint8_t avr2Status;
  int ret = EXIT_OK;

  command[0] = 0x05; /* User extended command group */
  command[1] = 0x82; /* I2CWrite command */
  command[2] = nBusIndex; /* I2C bus number */
  command[3] = nDevIndex; /* I2C device number */
  command[4] = nRegIndex; /* Address within device */
  command[5] = 1;
  command[6] = nData;
  expResponse[0] = 0x05; /* User extended command group */
  expResponse[1] = 0x02; /* I2CWrite response */

  memset(response, 0, sizeof(response));
  ret = avr2utilSendAvr2Command(
    pDevice,
    bVerbose,
    0,
    command,
    sizeof(command),
    expResponse,
    response,
    sizeof(response),
    &responseLength);
  if (EXIT_OK != ret) {
    goto done;
  }

  avr2Status = response[2];
  if (0 != avr2Status) {
    const TCHAR* pReason;

    switch (avr2Status) {
    case 1:
      pReason = _T("General failure");
      break;

    default:
      pReason = _T("** UNEXPECTED ERROR CODE **");
      break;
    }
    _tprintf(_T("*** AVR2 returned status 0x%02X => %s\n"), (unsigned int)response[2], pReason);
    ret = EXIT_AVR2_BAD_STATUS;
    goto done;
  }

done:
  return ret;
}

static int
setPage(
  const DeviceHandle* pDevice,
  bool_t bVerbose,
  uint8_t nBusIndex,
  uint8_t nDevIndex,
  uint8_t nPage,
  int* pnCurrentPage)
{
  int nCurrentPage = *pnCurrentPage;
  int ret = EXIT_OK;
  
  if (nCurrentPage < 0 || (uint8_t)nCurrentPage != nPage) {
    ret = writeRegInner(pDevice, bVerbose, nBusIndex, nDevIndex, SI5338_REGIDX_PAGE, nPage);
    if (EXIT_OK == ret) {
      *pnCurrentPage = (int)nPage;
    }
  }

  return ret;
}

static int
readReg(
  const DeviceHandle* pDevice,
  bool_t bVerbose,
  uint8_t nBusIndex,
  uint8_t nDevIndex,
  uint16_t nRegIndex,
  uint8_t* pnData,
  int* pnCurrentPage)
{
  uint8_t nPage = (nRegIndex >> 8) & 0x1;
  int ret;

  ret = setPage(pDevice, bVerbose, nBusIndex, nDevIndex, nPage, pnCurrentPage);
  if (EXIT_OK != ret) {
    return ret;
  }

  ret = readRegInner(pDevice, bVerbose, nBusIndex, nDevIndex, (uint8_t)nRegIndex, pnData);
  if (EXIT_OK != ret) {
    return ret;
  }

  return EXIT_OK;
}

static int
writeReg(
  const DeviceHandle* pDevice,
  bool_t bVerbose,
  uint8_t nBusIndex,
  uint8_t nDevIndex,
  uint16_t nRegIndex,
  uint8_t nData,
  int* pnCurrentPage)
{
  uint8_t nPage = (nRegIndex >> 8) & 0x1;
  int ret;

  ret = setPage(pDevice, bVerbose, nBusIndex, nDevIndex, nPage, pnCurrentPage);
  if (EXIT_OK != ret) {
    return ret;
  }

  ret = writeRegInner(pDevice, bVerbose, nBusIndex, nDevIndex, (uint8_t)nRegIndex, nData);
  if (EXIT_OK != ret) {
    return ret;
  }

  return EXIT_OK;
}

static int
writeRegMasked(
  const DeviceHandle* pDevice,
  bool_t bVerbose,
  uint8_t nBusIndex,
  uint8_t nDevIndex,
  uint16_t nRegIndex,
  uint8_t nMask,
  uint8_t nData,
  int* pnCurrentPage)
{
  int ret;
  uint8_t nOldData = 0, nNewData;

  if (nMask != 0xFF) {
    ret = readReg(pDevice, bVerbose, nBusIndex, nDevIndex, nRegIndex, &nOldData, pnCurrentPage);
    if (EXIT_OK != ret) {
      return ret;
    }
  }

  nNewData = (nData & nMask) | (nOldData & ~nMask); 

  ret = writeReg(pDevice, bVerbose, nBusIndex, nDevIndex, nRegIndex, nNewData, pnCurrentPage);
  if (EXIT_OK != ret) {
    return ret;
  }

  return EXIT_OK;
}

static int
configureSI5338Full(
  const DeviceHandle* pDevice,
  bool_t bVerbose,
  uint8_t nBusIndex,
  uint8_t nDevIndex,
  const SI5338RegMap* pRegMap)
{
  const unsigned int pllPollIntervalMs = 25; /* 25 ms poll interval waiting for PLL lock */
  const unsigned int maxCount = 10;          /* Timeout is 10 periods of poll interval waiting for PLL lock */
  char revision, devGrade;
  uint32_t nvmCode;
  uint8_t basePart, mask8, val8, fcal0, fcal1, fcal2;
  unsigned int i, count;
  int nCurrentPage = -1; /* Page register initially not established */
  int ret;

  /* Figure out part number */
  ret = readReg(pDevice, bVerbose, nBusIndex, nDevIndex, SI5338_REGIDX_REV_ID, &val8, &nCurrentPage);
  if (EXIT_OK != ret) {
    return ret;
  }
  revision = (char)('A' + (char)(val8 & 0x7U));

  ret = readReg(pDevice, bVerbose, nBusIndex, nDevIndex, SI5338_REGIDX_DEV_CONFIG2, &val8, &nCurrentPage);
  if (EXIT_OK != ret) {
    return ret;
  }
  basePart = (uint8_t)(val8 & 0x3FU);

  ret = readReg(pDevice, bVerbose, nBusIndex, nDevIndex, SI5338_REGIDX_DEV_CONFIG3, &val8, &nCurrentPage);
  if (EXIT_OK != ret) {
    return ret;
  }
  devGrade = (char)('A' - 1 + (char)((val8 & 0xF8U) >> 3));
  nvmCode = (uint32_t)((uint32_t)(val8 & 0x1) << 16);

  ret =  readReg(pDevice, bVerbose, nBusIndex, nDevIndex, SI5338_REGIDX_DEV_CONFIG4, &val8, &nCurrentPage);
  if (EXIT_OK != ret) {
    return ret;
  }
  nvmCode |= (uint32_t)((uint32_t)val8 << 0);

  ret =  readReg(pDevice, bVerbose, nBusIndex, nDevIndex, SI5338_REGIDX_DEV_CONFIG5, &val8, &nCurrentPage);
  if (EXIT_OK != ret) {
    return ret;
  }
  nvmCode |= (uint32_t)((uint32_t)val8 << 8);

  if (devGrade < 'A' || devGrade > 'B') {
  	_tprintf(_T("*** Unsupported part SI53%u%c-%c%lu\n"), (unsigned int)basePart, devGrade, (unsigned int)revision, (unsigned long)nvmCode);
    return EXIT_NOT_SI5338;
  } else {
  	_tprintf(_T("Part is SI53%u%c-%c%lu\n"), (unsigned int)basePart, devGrade, (unsigned int)revision, (unsigned long)nvmCode);
  }

  /* Disable all outputs */
  val8 = mask8 = 0x10U;
  ret = writeRegMasked(pDevice, bVerbose, nBusIndex, nDevIndex, SI5338_REGIDX_DRIVER_DISABLE, mask8, val8, &nCurrentPage);
  if (EXIT_OK != ret) {
    return ret;
  }

  /* Pause loss-of-lock detection */
  val8 = 0xE5U;
  mask8 = 0xFFU;
  ret = writeRegMasked(pDevice, bVerbose, nBusIndex, nDevIndex, SI5338_REGIDX_LOL_PAUSE, mask8, val8, &nCurrentPage);
  if (EXIT_OK != ret) {
    return ret;
  }

  /* Process SI5338 configuration records */
  for (i = 0; i < pRegMap->nEntryCount; i++) {
    uint16_t nRegAddr = (uint16_t)pRegMap->entries[i].nRegIndexLo + ((uint16_t)pRegMap->entries[i].nRegIndexHi << 8);

    ret = writeRegMasked(pDevice, bVerbose, nBusIndex, nDevIndex, nRegAddr,
      pRegMap->entries[i].nMaskValue, pRegMap->entries[i].nRegValue, &nCurrentPage);
    if (EXIT_OK != ret) {
      return ret;
    }
  }

  if (bVerbose) {
    _tprintf(_T("Wrote %u register map entries to SI5338\n"), i);
  }

  /* Validate input clock status */
  for (count = 0;; count++) {
    /* Check for input clock valid */
    ret = readReg(pDevice, bVerbose, nBusIndex, nDevIndex, SI5338_REGIDX_LOL_STATUS, &val8, &nCurrentPage);
    if (EXIT_OK != ret) {
      return ret;
    }

    if ((val8 & 0x04U) == 0U) {
      /* Input clock is valid */
      break;
    }
    if (count > maxCount) {
      break;
    }

    sleepMs(pllPollIntervalMs);
  }

  if (count > maxCount) {
    /* Timeout waiting for input clock valid */
    _tprintf(_T("*** Timeout waiting for input clock valid\n"));
    return EXIT_SI5338_CONFIG_FAILED;
  }

  /* Configure PLL for locking */
  val8 = 0x00U;
  mask8 = 0x80U;
  ret = writeRegMasked(pDevice, bVerbose, nBusIndex, nDevIndex, SI5338_REGIDX_FCAL_OVRD_EN, mask8, val8, &nCurrentPage);
  if (EXIT_OK != ret) {
    return ret;
  }

  /* Initiate locking of PLL */
  val8 = 0x02U;
  mask8 = 0xFFU; /* SI5338 AN411 says don't use read-modify-write for this register */
  ret = writeRegMasked(pDevice, bVerbose, nBusIndex, nDevIndex, SI5338_REGIDX_PLL_SOFT_RESET, mask8, val8, &nCurrentPage);
  if (EXIT_OK != ret) {
    return ret;
  }

  /* Resume loss-of-lock detection */
  val8 = 0x65U;
  mask8 = 0xFFU;
  ret = writeRegMasked(pDevice, bVerbose, nBusIndex, nDevIndex, SI5338_REGIDX_LOL_PAUSE, mask8, val8, &nCurrentPage);
  if (EXIT_OK != ret) {
    return ret;
  }

  for (count = 0; count < maxCount; count++) {
    /* SI5338 datasheet says wait 25 ms */
    sleepMs(pllPollIntervalMs);

    /* Check for PLL lock */
    ret = readReg(pDevice, bVerbose, nBusIndex, nDevIndex, SI5338_REGIDX_LOL_STATUS, &val8, &nCurrentPage);
    if (EXIT_OK != ret) {
      return ret;
    }

    if ((val8 & 0x15U) == 0U) {
      /* PLL locked, input clock valid, device calibration not in progress */
      break;
    }

    if (bVerbose) {
      _tprintf(_T("LOL_STATUS=0x%02x\n"), val8);
    }
  }

  if (count == maxCount) {
    /* Timeout waiting for PLL lock */
    _tprintf(_T("*** Timeout waiting for PLL lock; LOL_STATUS=0x%02x\n"), val8);
    return EXIT_SI5338_CONFIG_FAILED;
  }

  /* Copy FCAL registers to FCAL_OVRD registers */
  ret = readReg(pDevice, bVerbose, nBusIndex, nDevIndex, SI5338_REGIDX_FCAL2, &fcal2, &nCurrentPage);
  if (EXIT_OK != ret) {
    return ret;
  }
  ret = readReg(pDevice, bVerbose, nBusIndex, nDevIndex, SI5338_REGIDX_FCAL1, &fcal1, &nCurrentPage);
  if (EXIT_OK != ret) {
    return ret;
  }
  ret = readReg(pDevice, bVerbose, nBusIndex, nDevIndex, SI5338_REGIDX_FCAL0, &fcal0, &nCurrentPage);
  if (EXIT_OK != ret) {
    return ret;
  }

  /* Datasheet says set bits 7:2 to 000101 */
  fcal2 = (uint8_t)(fcal2 & 0x03U);
  fcal2 |= 0x14U;
  mask8 = 0xFFU;
  ret = writeRegMasked(pDevice, bVerbose, nBusIndex, nDevIndex, SI5338_REGIDX_FCAL_OVRD2, mask8, fcal2, &nCurrentPage);
  if (EXIT_OK != ret) {
    return ret;
  }
  ret = writeRegMasked(pDevice, bVerbose, nBusIndex, nDevIndex, SI5338_REGIDX_FCAL_OVRD1, mask8, fcal1, &nCurrentPage);
  if (EXIT_OK != ret) {
    return ret;
  }
  ret = writeRegMasked(pDevice, bVerbose, nBusIndex, nDevIndex, SI5338_REGIDX_FCAL_OVRD0, mask8, fcal0, &nCurrentPage);
  if (EXIT_OK != ret) {
    return ret;
  }

  /* Set FCAL_OVRD_EN enable bit */
  val8 = mask8 = 0x80U;
  ret = writeRegMasked(pDevice, bVerbose, nBusIndex, nDevIndex, SI5338_REGIDX_FCAL_OVRD_EN, mask8, val8, &nCurrentPage);
  if (EXIT_OK != ret) {
    return ret;
  }

  /* Reenable all outputs */
  mask8 = 0x10U;
  val8 = 0x00U;
  ret = writeRegMasked(pDevice, bVerbose, nBusIndex, nDevIndex, SI5338_REGIDX_DRIVER_DISABLE, mask8, val8, &nCurrentPage);
  if (EXIT_OK != ret) {
    return ret;
  }

  _tprintf(_T("SI5338 configuration apparently successful.\n"));

  return EXIT_OK;
}

extern int
avr2utilConfigSi5338Command(
  const DeviceHandle* pDevice,
  bool_t bVerbose,
  unsigned int nBusIndex,
  unsigned int nDevIndex,
  const TCHAR* pszRegMapPath)
{
  SI5338RegMap* pRegMap = NULL;
  SI5338RegMapError nRegMapError;
  int nExitCode = EXIT_OK;

  if (nBusIndex > 0xFFU) {
    _tprintf(_T("*** I2C bus number be in the range 0 to 255.\n"));
    return EXIT_INVALID_I2C_BUS;
  }

  if (nDevIndex > 0x7FU) {
    _tprintf(_T("*** I2C device number must be in the range 0 to 127.\n"));
    return EXIT_INVALID_I2C_DEVICE;
  }

  if (avr2utilIsInServiceMode(pDevice)) {
    _tprintf(_T("*** The config-si5338 command requires the device to not be in Service Mode.\n"));
    _tprintf(_T("    Please exit Service Mode manually using switch setting or by issuing the exit-service-mode command.\n"));
    return EXIT_WRONG_MODE;
  }

  nRegMapError = SI5338RegMapRead(pszRegMapPath, messageCallbackRoutine, &pRegMap);
  if (SI5338RegMapOk != nRegMapError) {
    const TCHAR* pszReason = SI5338RegMapGetErrorDesc(nRegMapError);

    _tprintf(_T("*** Failed to read file '%s' in SI5338 register map format: %s (error %d)\n"), pszRegMapPath, pszReason, nRegMapError);

    return EXIT_READ_REG_MAP_FAILED;
  }

  _tprintf(_T("File '%s' is an SI5338 register map with %u register / value entries(s)\n"), pszRegMapPath, pRegMap->nEntryCount);

  if (bVerbose) {
    dumpRegMap(pRegMap);
  }

  nExitCode = configureSI5338Full(pDevice, bVerbose, (uint8_t)nBusIndex, (uint8_t)nDevIndex, pRegMap);

  SI5338RegMapDestroy(pRegMap);

  return nExitCode;
}
