/*
** si5338_regmap.c - Functions for reading and writing SI5338A/B register map
**                   (.txt) files.
**
** Such files consist of a number of lines of the form
**
**   <register index>,<value>h
**
** where:
**
**   <register index> is a base 10 register index
**   <value> is a base 16 register value
**
** The SILabs ClockBuilder tool produces a register map file whose lines are
** in the above format.
**
** We also accept the # character as the beginning of a comment that continues until the end of line.
**
** (C) Copyright 2024 Alpha Data
*/

#if defined(_WIN32)

/* Windows */
# include <windows.h>
# include <tchar.h>
# define _stprintf_s0(buf, siz, fmt) _stprintf_s(buf, siz, fmt)
# define _stprintf_s1(buf, siz, fmt, arg1) _stprintf_s(buf, siz, fmt, arg1)
# define _stprintf_s2(buf, siz, fmt, arg1, arg2) _stprintf_s(buf, siz, fmt, arg1, arg2)
# define _stprintf_s3(buf, siz, fmt, arg1, arg2, arg3) _stprintf_s(buf, siz, fmt, arg1, arg2, arg3)

#else

/* Linux or VxWorks */
# define _T(x) x
# define _TEOF EOF
# define _TINT char
# define _tprintf printf
# define _tcslen strlen
# define _fgettc fgetc
# define _istcntrl iscntrl
# define _istdigit isdigit
# define _istprint isprint
# define _istspace isspace
# define _totlower tolower
# define _stscanf_s sscanf
# define _stprintf_s0(buf, siz, fmt) sprintf(buf, fmt)
# define _stprintf_s1(buf, siz, fmt, arg1) sprintf(buf, fmt, arg1)
# define _stprintf_s2(buf, siz, fmt, arg1, arg2) sprintf(buf, fmt, arg1, arg2)
# define _stprintf_s3(buf, siz, fmt, arg1, arg2, arg3) sprintf(buf, fmt, arg1, arg2, arg3)
# define FALSE (0)
# define TRUE (1)

#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <errno.h>
#include <ctype.h>

#include <sdk_common.h>
#include <platform_endian.h>

#include "si5338_regmap.h"

#define ARRAY_LENGTH(x) (sizeof(x) / sizeof((x)[0]))

#define OFFSET_OF(t, f) ((size_t)((uintptr_t)&(((t*)0)->f) - (uintptr_t)((t*)0)))

#define MAX_LINE_LENGTH (32) /* Maximum line length excluding CR / LF, excluding NUL terminator */

#define MAX_MESSAGE_LENGTH (255) /* Maximum number of characters in a message, excluding NUL terminator */

#define MAX_BAD_LINE_MESSAGE (10)  /* Maximum number of messages about malformed lines to output */
#define MAX_BAD_VALUE_MESSAGE (10) /* Maximum number of messages about bad values to output */

/* From SI Labs documentation */
static const uint8_t g_registerMasks[351] = {
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x1D, 0x00,  /* 0x000 */
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,  /* 0x008 */
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,  /* 0x010 */
  0x00, 0x00, 0x00, 0x80, 0xFF, 0xFF, 0xFF, 0xFF,  /* 0x018 */
  0xFF, 0xFF, 0xFF, 0xFF, 0x1F, 0x1F, 0x1F, 0x1F,  /* 0x020 */
  0xFF, 0x7F, 0x3F, 0x00, 0x00, 0xFF, 0xFF, 0x3F,  /* 0x028 */
  0xFF, 0xFF, 0xFF, 0xFF, 0x7F, 0xFF, 0xFF, 0xFF,  /* 0x030 */
  0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0x3F, 0x7F,  /* 0x038 */
  0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,  /* 0x040 */
  0xFF, 0x3F, 0x7F, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,  /* 0x048 */
  0xFF, 0xFF, 0xFF, 0xFF, 0x3F, 0x7F, 0xFF, 0xFF,  /* 0x050 */
  0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0x3F,  /* 0x058 */
  0x00, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,  /* 0x060 */
  0xFF, 0xFF, 0xBF, 0xFF, 0x7F, 0xFF, 0xFF, 0xFF,  /* 0x068 */
  0x7F, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,  /* 0x070 */
  0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,  /* 0x078 */
  0xFF, 0x0F, 0x0F, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,  /* 0x080 */
  0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,  /* 0x088 */
  0xFF, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,  /* 0x090 */
  0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0x0F, 0x0F,  /* 0x098 */
  0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,  /* 0x0A0 */
  0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,  /* 0x0A8 */
  0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0x0F, 0xFF, 0xFF,  /* 0x0B0 */
  0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,  /* 0x0B8 */
  0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,  /* 0x0C0 */
  0xFF, 0xFF, 0xFF, 0x0F, 0xFF, 0xFF, 0xFF, 0xFF,  /* 0x0C8 */
  0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,  /* 0x0D0 */
  0xFF, 0xFF, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,  /* 0x0D8 */
  0x00, 0x00, 0x04, 0x00, 0x00, 0x00, 0x0F, 0x00,  /* 0x0E0 */
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,  /* 0x0E8 */
  0x00, 0x0F, 0x02, 0x00, 0x00, 0x00, 0x0F, 0x00,  /* 0x0F0 */
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xFF,  /* 0x0F8 */
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,  /* 0x100 */
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,  /* 0x108 */
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,  /* 0x110 */
  0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0xFF,  /* 0x118 */
  0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,  /* 0x120 */
  0xFF, 0xFF, 0xFF, 0x0F, 0x00, 0x00, 0x00, 0xFF,  /* 0x128 */
  0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,  /* 0x130 */
  0xFF, 0xFF, 0xFF, 0x0F, 0x00, 0x00, 0x00, 0xFF,  /* 0x138 */
  0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,  /* 0x140 */
  0xFF, 0xFF, 0xFF, 0x0F, 0x00, 0x00, 0x00, 0xFF,  /* 0x148 */
  0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF, 0xFF,  /* 0x150 */
  0xFF, 0xFF, 0xFF, 0x0F, 0x00, 0x00, 0x00         /* 0x158 */
};

static SI5338RegMapError
readLine(
  FILE* pFile,
  TCHAR* pLineBuffer,
  unsigned int* pLength,
  bool_t *pbBadCharacter)
{
  unsigned int pos = 0;
  SI5338RegMapError ret = SI5338RegMapOk;
  bool_t bBadCharacter = FALSE;
  bool_t bInComment = FALSE;

  while (!feof(pFile)) {
    _TINT c;
    
    c = _fgettc(pFile);
    if (c == _TEOF) {
      if (ferror(pFile)) {
        ret = SI5338RegMapReadError;
      }
      break;
    } else if (c == _T('\n')) {
      break;
    } else if (c == _T('\r')) {
      /* Skip CR characters */
    } else {
      if (_istprint(c) || _istspace(c)) {
        if (c == _T('#')) {
          /* # denotes beginning of comment until end of line */
          bInComment = TRUE;
        }
        if (!bInComment && pos < MAX_LINE_LENGTH) {
          pLineBuffer[pos++] = (TCHAR)c;
        }
      } else {
        /* Skip other nonprintable characters and record as warning */
        bBadCharacter = TRUE;
      }
    }
  }

  pLineBuffer[pos] = _T('\0');

  *pLength = pos;
  *pbBadCharacter = bBadCharacter;

  return ret;
}

#define NUM_TOKEN_EXPECTED 2
#define MAX_TOKEN_LENGTH 16

static bool_t parseLine(
  const TCHAR* pLine,
  unsigned int* pRegIndex,
  unsigned int* pRegValue)
{
  unsigned int i, nTokenCount = 0;
  unsigned int nLinePos = 0, nTokenPos;
  TCHAR tokenBuf[NUM_TOKEN_EXPECTED][MAX_TOKEN_LENGTH], c;
  bool_t bEndOfLine = FALSE;
  unsigned int nTokenValue[NUM_TOKEN_EXPECTED];
  int nConverted;

  while (nTokenCount < NUM_TOKEN_EXPECTED) {
    /* Skip whitespace */
    while (!bEndOfLine) {
      c = pLine[nLinePos];
      if (_T('\0') == c) {
        bEndOfLine = TRUE;
      } else if (!_istspace(c)) {
        break;
      } else {
        nLinePos++;
      }
    }

    if (bEndOfLine) {
      break;
    }

    /* Accumulate token characters */
    nTokenPos = 0;
    while (!bEndOfLine) {
      c = pLine[nLinePos];
      if (_T('\0') == c) {
        bEndOfLine = TRUE;
      } else if (_istspace(c)) {
        break;
      } else if (c == _T(',')) {
        nLinePos++;
        break;
      } else {
        if (nTokenPos < MAX_TOKEN_LENGTH - 1) {
          tokenBuf[nTokenCount][nTokenPos] = _totlower(c);
          nTokenPos++;
        }

        nLinePos++;
      }
    }
    tokenBuf[nTokenCount][nTokenPos] = _T('\0');
    nTokenCount++;

    if (bEndOfLine) {
      break;
    }
  }

  if (nTokenCount > 0 && nTokenCount < NUM_TOKEN_EXPECTED) {
    return FALSE; /* Bad line - Does not have 2 tokens */
  }

  for (i = 0; i < nTokenCount; i++) {
    size_t nTokenLength = _tcslen(tokenBuf[i]);
    bool_t bTokenIsHex = tokenBuf[i][nTokenLength - 1] == _T('h') ? TRUE : FALSE;

    if (bTokenIsHex) {
      tokenBuf[i][nTokenLength - 1] = _T('\0'); /* Remove 'h' from end of token */
      nConverted = _stscanf_s(tokenBuf[i], _T("%x"), &nTokenValue[i]);
    } else {
      nConverted = _stscanf_s(tokenBuf[i], _T("%u"), &nTokenValue[i]);
    }

    if (0 == nConverted) {
      /* Bad line - Token isn't a number */
      return FALSE;
    }
  }

  /* Parsed line OK */
  *pRegIndex = nTokenValue[0];
  *pRegValue = nTokenValue[1];

  return TRUE;
}

SI5338RegMapError SI5338RegMapReadFromFile(
  FILE* pFile,
  SI5338RegMapMessageCallback* pMessageCallback,
  SI5338RegMap** ppRegMap)
{
  SI5338RegMapError eResult = SI5338RegMapOk;
  unsigned int nPass, nLineNumber;
  unsigned int nEntryCount = 0, nEntryCountAlloc = 0;
  unsigned int nBadLineCount, nBadValueCount;
  SI5338RegMap* pRegMap = NULL;
  TCHAR messageBuf[MAX_MESSAGE_LENGTH + 1];
  TCHAR lineBuf[MAX_LINE_LENGTH + 1];
  unsigned int nLineLength;
  bool_t bBadCharacter;

  for (nPass = 0; nPass < 2; nPass++) {
    if (fseek(pFile, 0, SEEK_SET)) {
      if (NULL != pMessageCallback) {
        _stprintf_s1(messageBuf, ARRAY_LENGTH(messageBuf), _T("Failed to seek to beginning of input file; errno=%d"), errno);
        pMessageCallback(messageBuf, SI5338RegMapMessageError);
      }

      eResult = SI5338RegMapSeekError;
      goto out;
    }

    nEntryCount = 0;
    nLineNumber = 1;
    nBadLineCount = 0;
    nBadValueCount = 0;

    while (!feof(pFile)) {
      unsigned int nRegIndex, nRegValue;

      eResult = readLine(pFile, lineBuf, &nLineLength, &bBadCharacter);
      if (SI5338RegMapOk != eResult) {
        goto out;
      }

      if (0 == nLineLength) {
        continue;
      }

      if (parseLine(lineBuf, &nRegIndex, &nRegValue)) {
        bool_t bValuesOK = TRUE;

        if (nRegIndex >= ARRAY_LENGTH(g_registerMasks)) {
          bValuesOK = FALSE;

          if (nBadValueCount < MAX_BAD_VALUE_MESSAGE && 1 == nPass) {
            _stprintf_s2(messageBuf, ARRAY_LENGTH(messageBuf), _T("Out of range register index LS value 0x%X on line %u - ignoring."), nRegIndex, nLineNumber);
            pMessageCallback(messageBuf, SI5338RegMapMessageWarning);

            if (nBadValueCount == MAX_BAD_VALUE_MESSAGE - 1) {
              _stprintf_s0(messageBuf, ARRAY_LENGTH(messageBuf), _T("NOTE: Further messages about out of range values will be suppressed"));
              pMessageCallback(messageBuf, SI5338RegMapMessageInfo);
            }

            nBadValueCount++;
          }
        }

        if (nRegValue > 0xFF) {
          bValuesOK = FALSE;

          if (nBadValueCount < MAX_BAD_VALUE_MESSAGE && 1 == nPass) {
            _stprintf_s2(messageBuf, ARRAY_LENGTH(messageBuf), _T("Out of range register value 0x%X on line %u - ignoring."), nRegValue, nLineNumber);
            pMessageCallback(messageBuf, SI5338RegMapMessageWarning);

            if (nBadValueCount == MAX_BAD_VALUE_MESSAGE - 1) {
              _stprintf_s0(messageBuf, ARRAY_LENGTH(messageBuf), _T("NOTE: Further messages about out of range values will be suppressed"));
              pMessageCallback(messageBuf, SI5338RegMapMessageInfo);
            }

            nBadValueCount++;
          }
        }

        if (bValuesOK) {
          if (1 == nPass && nEntryCount < nEntryCountAlloc) { /* Check in case file changes somehow between passes */
            pRegMap->entries[nEntryCount].nRegIndexLo = (uint8_t)(nRegIndex & 0xFF);
            pRegMap->entries[nEntryCount].nRegIndexHi = (uint8_t)((nRegIndex >> 8) & 0xFF);
            pRegMap->entries[nEntryCount].nMaskValue = g_registerMasks[nRegIndex];
            pRegMap->entries[nEntryCount].nRegValue = (uint8_t)nRegValue;
          }

          nEntryCount++;
        }
      } else {
        if (NULL != pMessageCallback && 1 == nPass) {
          if (nBadLineCount < MAX_BAD_LINE_MESSAGE) {
            _stprintf_s1(messageBuf, ARRAY_LENGTH(messageBuf), _T("Line %u of input file appears to be malformed - ignoring."), nLineNumber);
            pMessageCallback(messageBuf, SI5338RegMapMessageWarning);

            if (nBadLineCount == MAX_BAD_LINE_MESSAGE - 1) {
              _stprintf_s0(messageBuf, ARRAY_LENGTH(messageBuf), _T("NOTE: Further messages about malformed lines will be suppressed"));
              pMessageCallback(messageBuf, SI5338RegMapMessageInfo);
            }
          }
        }

        nBadLineCount++;
      }

      nLineNumber++;
    }

    if (0 == nEntryCount) {
      eResult = SI5338RegMapBadFile;
      goto out;
    }

    if (0 == nPass) {
      size_t nBytesAlloc;

      nEntryCountAlloc = nEntryCount;
      if (nEntryCountAlloc > SI5338_MAX_ENTRY_COUNT) {
        if (NULL != pMessageCallback) {
          _stprintf_s2(
              messageBuf, ARRAY_LENGTH(messageBuf),
              _T("More than %u register / value pairs in input file - ignoring excess %u pair(s)."),
              SI5338_MAX_ENTRY_COUNT, nEntryCount - SI5338_MAX_ENTRY_COUNT);
          pMessageCallback(messageBuf, SI5338RegMapMessageWarning);
        }

        nEntryCountAlloc = SI5338_MAX_ENTRY_COUNT;
      }

      nBytesAlloc = sizeof(pRegMap->nEntryCount) + nEntryCountAlloc * sizeof(pRegMap->entries[0]);
      pRegMap = (SI5338RegMap*)malloc(nBytesAlloc);
      if (NULL == pRegMap) {
        if (NULL != pMessageCallback) {
          _stprintf_s3(
            messageBuf, ARRAY_LENGTH(messageBuf),
            _T("Failed to allocate %llu (0x%llX) bytes for holding register map data; errno=%d"),
            (unsigned long long)nBytesAlloc, (unsigned long long)nBytesAlloc, errno);
          pMessageCallback(messageBuf, SI5338RegMapMessageError);
        }

        return SI5338RegMapAllocationError;
      }

      pRegMap->nEntryCount = (uint8_t)nEntryCountAlloc;
    }
  }

out:
  if (SI5338RegMapOk == eResult) {
    *ppRegMap = pRegMap;
  } else {
    if (NULL != pRegMap) {
      free(pRegMap);
    }
  }

  return eResult;
}

SI5338RegMapError SI5338RegMapRead(
  const TCHAR* pPath,
  SI5338RegMapMessageCallback* pMessageCallback,
  SI5338RegMap** ppRegMap)
{
  SI5338RegMapError ret = SI5338RegMapOk;
  TCHAR messageBuf[MAX_MESSAGE_LENGTH + 1];
  FILE* pFile = NULL;
  int err;

#ifdef _MSC_VER
  err = _tfopen_s(&pFile, pPath, _T("rt"));
#else
# ifdef _WIN32
  pFile = _tfopen(pPath, "rb");
# else
  pFile = fopen(pPath, "r");
# endif
  err = (NULL == pFile) ? errno : 0;
#endif
  if (err) {
    switch (err) {
    case EACCES:
      return SI5338RegMapPermissionDenied;

    case ENOENT:
      return SI5338RegMapNotFound;

    default:
      if (NULL != pMessageCallback) {
        _stprintf_s1(messageBuf, ARRAY_LENGTH(messageBuf), _T("Failed to seek to beginning of input file; errno=%d"), errno);
        pMessageCallback(messageBuf, SI5338RegMapMessageError);
      }

      return SI5338RegMapOpenFailed;
    }
  }

  ret = SI5338RegMapReadFromFile(pFile, pMessageCallback, ppRegMap);

  fclose(pFile);

  return ret;
}

void SI5338RegMapDestroy(
  SI5338RegMap* pRegMap)
{
  free(pRegMap);
}

const TCHAR* SI5338RegMapGetErrorDesc(
  SI5338RegMapError nError)
{
  const TCHAR* pszReason;

  switch (nError) {
  case SI5338RegMapOk:
    pszReason = _T("no error");
    break;

  case SI5338RegMapBadFile:
    pszReason = _T("bad register map file");
    break;

  case SI5338RegMapAllocationError:
    pszReason = _T("unable to allocate memory");
    break;

  case SI5338RegMapPermissionDenied:
    pszReason = _T("access denied");
    break;

  case SI5338RegMapNotFound:
    pszReason = _T("file not found");
    break;

  case SI5338RegMapOpenFailed:
    pszReason = _T("unexpected error opening file");
    break;

  case SI5338RegMapSeekError:
    pszReason = _T("seek error while reading file");
    break;

  case SI5338RegMapReadError:
    pszReason = _T("error while reading file");
    break;

  default:
    pszReason = _T("unknown error");
    break;
  }

  return pszReason;
}
